/***************************************************************************/
/*                                                                         */
/*  ttgxvar.c                                                              */
/*                                                                         */
/*    TrueType GX Font Variation loader                                    */
/*                                                                         */
/*  Copyright 2004 by                                                      */
/*  David Turner, Robert Wilhelm, Werner Lemberg and George Williams.      */
/*                                                                         */
/*  This file is part of the FreeType project, and may only be used,       */
/*  modified, and distributed under the terms of the FreeType project      */
/*  license, LICENSE.TXT.  By continuing to use, modify, or distribute     */
/*  this file you indicate that you have read the license and              */
/*  understand and accept it fully.                                        */
/*                                                                         */
/***************************************************************************/

/***************************************************************************/
/*                                                                         */
/* Apple documents these tables at                                         */
/*    http:// developer.apple.com/fonts/TTRefMan/RM06/Chap6[fgca]var.html  */
/* The documentation for fvar is inconsistant. At one point it says that   */
/*  countSizePairs should be 3, at another point 2. It should be 2.        */
/* The documentation for gvar is not intelligable, cvar refers you to gvar */
/*  and is thus also incomprehensible                                      */
/* The documentation for avar appears correct, but Apple has no fonts      */
/*  with an 'avar' table, so it's hard to test.                            */
/* I (GWW) got a lot of help from John Jenkins @ Apple in figuring this out*/
/*                                                                         */
/***************************************************************************/

#ifdef TT_CONFIG_OPTION_GX_VAR_SUPPORT

#include <ft2build.h>
#include FT_INTERNAL_DEBUG_H
#include FT_CONFIG_CONFIG_H
#include FT_INTERNAL_STREAM_H
#include FT_INTERNAL_SFNT_H
#include FT_TRUETYPE_IDS_H
#include FT_MULTIPLE_MASTERS_H

#include "ttdriver.h"
#include "ttgxvar.h"

#include "tterrors.h"

/* I'm sure this is the wrong approach, but someone who knows freetype better */
/*  than I should fix it, I couldn't figure out what I was supposed to do     */
#define FT_Stream_FTell( stream ) ( (stream)->cursor - (stream)->base )
#define FT_Stream_SeekSet( stream, off ) \
	( (stream)->cursor = (stream)->base+(off) )


  /*************************************************************************/
  /*                                                                       */
  /* The macro FT_COMPONENT is used in trace mode.  It is an implicit      */
  /* parameter of the FT_TRACE() and FT_ERROR() macros, used to print/log  */
  /* messages during execution.                                            */
  /*                                                                       */
#undef  FT_COMPONENT
#define FT_COMPONENT  trace_ttgxvar


  /*************************************************************************/
  /*************************************************************************/
  /*****                                                               *****/
  /*****                       Internal Routines                       *****/
  /*****                                                               *****/
  /*************************************************************************/
  /*************************************************************************/


  /*************************************************************************/
  /*                                                                       */
  /* The macro ALL_POINTS is used in by ft_var_readpackedpoints, it        */
  /* indicates that there is a delta for every point without needing to    */
  /* enumerate all of them.                                                */
  /*                                                                       */
#define ALL_POINTS	(FT_UShort *) (-1)


  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    ft_var_readpackedpoints                                            */
  /*                                                                       */
  /* <Description>                                                         */
  /*    read a set of points to which the following deltas will apply      */
  /*    points are packed with a run length encoding                       */
  /*                                                                       */
  /* <Input>                                                               */
  /*    stream       :: the data stream.                                   */
  /*                                                                       */
  /* <Output>                                                              */
  /*    point_cnt    :: the number of points read. 0=> all points in       */
  /*                    the glyph will be affected, without enumerating    */
  /*                    them individually                                  */
  /*                                                                       */
  /* <Return>                                                              */
  /*    An array of FT_UShort containing the affected points               */
  /*    or the special value ALL_POINTS                                    */
  /*                                                                       */
  static FT_UShort *
  ft_var_readpackedpoints(FT_Stream stream,
                          FT_Int   *point_cnt)
  {
    FT_UShort *points;
    FT_Int    n;
    FT_Int    runcnt;
    FT_Int    i;
    FT_Int    j;
    FT_Int    first;
    FT_Memory memory = stream->memory;
    FT_Error  error;

    *point_cnt = n = FT_GET_BYTE();
    if ( n==0 )
  return( ALL_POINTS );

    if ( n&0x80 )
	n = FT_GET_BYTE()|((n&0x7f)<<8);

    if ( FT_NEW_ARRAY(points,n))
  return( NULL );

    i = 0;
    while ( i<n ) {
	runcnt = FT_GET_BYTE();
	if ( runcnt&0x80 )
	{
	    runcnt = (runcnt&0x7f);
	    points[i++] = first = FT_GET_USHORT();
	    /* first point not included in runcount */
	    for ( j=0; j<runcnt; ++j )
		points[i++] = (first += FT_GET_USHORT());
	}
	else
	{
	    points[i++] = first = FT_GET_BYTE();
	    for ( j=0; j<runcnt; ++j )
		points[i++] = (first += FT_GET_BYTE());
	}
    }
  return( points );
  }



  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    ft_var_readpackeddeltas                                            */
  /*                                                                       */
  /* <Description>                                                         */
  /*    read a set of deltas. These are packed slightly differently        */
  /*    than points. In particular there is no overall count               */
  /*                                                                       */
  /* <Input>                                                               */
  /*    stream       :: the data stream.                                   */
  /*                                                                       */
  /*    delta_cnt    :: the number of to be read                           */
  /*                                                                       */
  /* <Return>                                                              */
  /*    An array of FT_Short containing the deltas for the affected        */
  /*    points. (this only gets the deltas for one dimension. it will      */
  /*    generally be called twice, once for x, once for y. When used in    */
  /*    cvt table, it will only be called once).                           */
  /*                                                                       */
  static FT_Short *
  ft_var_readpackeddeltas(FT_Stream stream,
                          FT_Int    delta_cnt)
  {
    FT_Short *deltas;
    FT_Int    runcnt;
    FT_Int    i;
    FT_Int    j;
    FT_Memory memory = stream->memory;
    FT_Error  error;

    if ( FT_NEW_ARRAY(deltas,delta_cnt))
  return( NULL );

    i = 0;
    while ( i<delta_cnt ) {
	runcnt = FT_GET_BYTE();
	if ( runcnt&0x80 )
	{
	    /* runcnt zeros get added */
	    for ( j=0; j<=(runcnt&0x3f) && i<delta_cnt ; ++j )
		deltas[i++] = 0;
	}
	else if ( runcnt&0x40 )
	{
	    /* runcnt shorts from the stack */
	    for ( j=0 ; j<=(runcnt&0x3f) && i<delta_cnt ; ++j )
		deltas[i++] = FT_GET_SHORT();
	}
	else
	{
	    /* runcnt signed bytes from the stack */
	    for ( j=0; j<=(runcnt&0x3f) && i<delta_cnt; ++j )
		deltas[i++] = FT_GET_CHAR();
	}
	if ( j<=(runcnt&0x3f) )
	{
	    /* Bad format */
	    FT_FREE(deltas);
  return( NULL );
	}
    }
  return( deltas );
  }


  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    ft_var_load_avar                                                   */
  /*                                                                       */
  /* <Description>                                                         */
  /*    parses the avar table if present. It need not be, so we never      */
  /*    return an error.                                                   */
  /*                                                                       */
  /* <InOut>                                                               */
  /*    face           :: the font.                                        */
  /*                                                                       */
  /* <Return>                                                              */
  /*    void.                                                              */
  /*                                                                       */
  static void
  ft_var_load_avar(TT_Face  face)
  {
    FT_Stream  stream = face->root.stream;
    FT_Memory  memory = stream->memory;
    GX_Blend   blend  = face->blend;
    FT_Error   error;
    FT_ULong   version;
    FT_Int     axis_cnt;
    FT_Int     i, j;
    FT_ULong   table_len;

    blend->avar_checked = TRUE;
    if ( (error = face->goto_table( face, TTAG_avar, stream, &table_len )) )
return;
    if ( FT_FRAME_ENTER( table_len ) )
return;

    version    = FT_GET_LONG();
    axis_cnt   = FT_GET_USHORT();
    if ( version!=0x00010000L || axis_cnt != blend->num_axis )
return;

    if ( FT_NEW_ARRAY( blend->avar, axis_cnt ))
      goto Exit;

    for ( i=0; i<axis_cnt; ++i )
    {
      blend->avar[i].num_points = FT_GET_USHORT();
      if ( FT_NEW_ARRAY(blend->avar[i].orig, blend->avar[i].num_points ) ||
           FT_NEW_ARRAY(blend->avar[i].final, blend->avar[i].num_points ) )
      {
        for ( j=0; j<=i; ++j )
	{
	  FT_FREE(blend->avar[i].orig);
	  FT_FREE(blend->avar[i].final);
	}
	FT_FREE(blend->avar);
	blend->avar = NULL;
	goto Exit;
      }
      for ( j=0; j<blend->avar[i].num_points; ++j )
      {
        blend->avar[i].orig[j]  = FT_GET_SHORT()<<2;    /* Convert to Fixed */
        blend->avar[i].final[j] = FT_GET_SHORT()<<2;    /* Convert to Fixed */
      }
    }

  Exit:
    FT_FRAME_EXIT();
  }


  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    ft_var_load_gvar                                                   */
  /*                                                                       */
  /* <Description>                                                         */
  /*    parses the avar table if present. It need not be, so we never      */
  /*    return an error.                                                   */
  /*                                                                       */
  /* <InOut>                                                               */
  /*    face           :: the font.                                        */
  /*                                                                       */
  /* <Return>                                                              */
  /*    FreeType error code.  0 means success.                             */
  /*                                                                       */
  static FT_Error
  ft_var_load_gvar(TT_Face  face)
  {
    FT_Stream  stream = face->root.stream;
    FT_Memory  memory = stream->memory;
    GX_Blend   blend  = face->blend;
    FT_Error   error;
    FT_ULong   version;
    FT_Int     axis_cnt;
    FT_Int     i, j;
    FT_ULong   table_len;
    FT_ULong   gvar_start;
    FT_ULong   coord_start, glyphs_start;
    FT_Int     flag;

    if ( (error = face->goto_table( face, TTAG_gvar, stream, &table_len )) )
      goto Exit;
    if ( FT_FRAME_ENTER( table_len ) )
      goto Exit;
    gvar_start = FT_Stream_FTell( stream );

    version           = FT_GET_LONG();
    axis_cnt          = FT_GET_USHORT();
    blend->tuplecount = FT_GET_USHORT();
    coord_start       = gvar_start + FT_GET_LONG();
    blend->gv_glyphcnt= FT_GET_USHORT();
    flag              = FT_GET_USHORT();
    glyphs_start      = gvar_start + FT_GET_LONG();
    
    if ( version!=0x00010000L || axis_cnt != blend->mmvar->num_axis )
    {
      error = TT_Err_Invalid_Table;
      goto FExit;
    }

    if ( FT_NEW_ARRAY( blend->glyphoffsets, blend->gv_glyphcnt+1 ) )
      goto FExit;

    if ( flag&1 )
    { /* Long offsets (one more offset than glyph, to mark size of last) */
      for ( i=0; i<=blend->gv_glyphcnt; ++i )
        blend->glyphoffsets[i] = glyphs_start + FT_GET_LONG();
    }
    else
    { /* Short offsets (one more offset than glyph, to mark size of last) */
      for ( i=0; i<=blend->gv_glyphcnt; ++i )
        blend->glyphoffsets[i] = glyphs_start + FT_GET_USHORT()*2;	/* Undocumented *2 */
    }

    if ( blend->tuplecount!=0 )
    {
      if ( FT_NEW_ARRAY( blend->tuplecoords, axis_cnt*blend->tuplecount ) )
        goto FExit;
      for ( i=0; i<blend->tuplecount; ++i )
        for ( j=0 ; j<axis_cnt; ++j )
          blend->tuplecoords[ i*axis_cnt + j ] = FT_GET_SHORT()<<2; /* convert to FT_Fixed */
    }

  FExit:
    FT_FRAME_EXIT();
  Exit:
    return error;
  }


  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    ft_var_apply_tuple                                                 */
  /*                                                                       */
  /* <Description>                                                         */
  /*    figure out whether a given tuple (design) applies to the current   */
  /*    blend, and if so, what is the scaling factor                       */
  /*                                                                       */
  /* <Input>                                                               */
  /*    blend           :: the current blend of the font.                  */
  /*                                                                       */
  /*    tupleIndex      :: flag saying whether this is an intermediate     */
  /*                       tuple or not.                                   */
  /*                                                                       */
  /*    tuple_coords    :: the coordiates of the tuple in normalized       */
  /*                       axis units.                                     */
  /*                                                                       */
  /*    im_start_coords :: the initial coords where this tuple starts      */
  /*                       to apply (for intermediate coords)              */
  /*                                                                       */
  /*    im_end_coords   :: the final coords after which this tuple no      */
  /*                       longer applies (for intermediate coords)        */
  /*                                                                       */
  /* <Return>                                                              */
  /*    A Fixed value containing the scaling factor.                       */
  /*                                                                       */
  static FT_Fixed
  ft_var_apply_tuple(GX_Blend  blend,
		     FT_UShort tupleIndex,
		     FT_Fixed *tuple_coords,
		     FT_Fixed *im_start_coords,
		     FT_Fixed *im_end_coords )
  {
    FT_Int   i;
    FT_Fixed apply;
    FT_Fixed temp;
		     
    apply = 0x10000L;
    for ( i=0; i<blend->num_axis; ++i )
    {
      if ( tuple_coords[i]==0 )
	/* I'm not sure why (for intermediate tuples) we don't check */
	/*  against start/end, but my docs say we don't, nor why we */
	/*  don't scale along the axis */
    continue;
      else if ( blend->normalizedcoords[i]==0 ||
          (blend->normalizedcoords[i]<0 && tuple_coords[i]>0) ||
          (blend->normalizedcoords[i]>0 && tuple_coords[i]<0))
      {
        apply = 0;
    break;
      }
      else if ( !(tupleIndex&0x4000 ) )
        /* Not an intermediate tuple */
        apply = FT_MulDiv(apply,
            (blend->normalizedcoords[i]>0 ?
                blend->normalizedcoords[i] :
                blend->normalizedcoords[i]),
            0x10000L );
      else if ( blend->normalizedcoords[i]<=im_start_coords[i] ||
                blend->normalizedcoords[i]>=im_end_coords[i] )
      {
        apply = 0;
    break;
      } 
      else if ( blend->normalizedcoords[i]<tuple_coords[i] )
      {
        temp = FT_MulDiv(
		    blend->normalizedcoords[i]-im_start_coords[i],
		    0x10000L,
		    tuple_coords[i]-im_start_coords[i]);
        apply = FT_MulDiv(apply,temp,0x10000L);
      }
      else
      {
        temp = FT_MulDiv(
		    im_end_coords[i]-blend->normalizedcoords[i],
		    0x10000L,
		    im_end_coords[i]-tuple_coords[i]);
        apply = FT_MulDiv(apply,temp,0x10000L);
      }
    }

  return apply;
  }


  /*************************************************************************/
  /*************************************************************************/
  /*****                                                               *****/
  /*****               MULTIPLE MASTERS SERVICE FUNCTIONS              *****/
  /*****                                                               *****/
  /*************************************************************************/
  /*************************************************************************/


  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    TT_Get_MM_Var                                                      */
  /*                                                                       */
  /* <Description>                                                         */
  /*    checks that the font's fvar table is valid, parses it,             */
  /*    and returns those data                                             */
  /*                                                                       */
  /* <InOut>                                                               */
  /*    face         :: the font.                                          */
  /*                    initialize the blend struct                        */
  /*                                                                       */
  /* <Output>                                                              */
  /*    master       :: return the fvar data (must be freed by caller)     */
  /*                                                                       */
  /* <Return>                                                              */
  /*    FreeType error code.  0 means success.                             */
  /*                                                                       */
  FT_LOCAL_DEF( FT_Error )
  TT_Get_MM_Var( TT_Face      face,
                 FT_MM_Var**  master )
  {
    FT_Stream  stream = face->root.stream;
    FT_Memory  memory = face->root.memory;
    FT_ULong   table_len;
    FT_Error   error;
    FT_Long    version;
    FT_ULong   data_off;
    FT_ULong   fvar_start;
    FT_Int     cnt;
    FT_Int     i,j;
    FT_Int     axis_cnt, axis_size;
    FT_Int     instance_cnt, instance_size;
    FT_MM_Var *mmvar;
    FT_Fixed  *next_coords;
    FT_String *next_name;
    FT_Var_Axis        *a;
    FT_Var_Named_Style *ns;

    if ( face->blend == NULL )
    {

      /* Both fvar and gvar must be present */
      if ( (error = face->goto_table( face, TTAG_gvar, stream, &table_len )) )
        goto Exit;
      if ( (error = face->goto_table( face, TTAG_fvar, stream, &table_len )) )
        goto Exit;
      if ( FT_FRAME_ENTER( table_len ) )
        goto Exit;

      fvar_start = FT_Stream_FTell( stream );
      version = FT_GET_LONG();
      data_off = FT_GET_USHORT();
      cnt = FT_GET_USHORT();
      axis_cnt = FT_GET_USHORT();
      axis_size = FT_GET_USHORT();
      instance_cnt = FT_GET_USHORT();
      instance_size = FT_GET_USHORT();

      if ( version != 0x00010000L || cnt != 2 || axis_size!=20 ||
	      instance_size != 4+4*axis_cnt ||
	      data_off + axis_cnt*20 + instance_cnt*instance_size > table_len )
      {
        error = TT_Err_Invalid_Table;
        goto FExit;
      }

      if ( FT_ALLOC( face->blend, sizeof(GX_BlendRec)))
        goto FExit;
      face->blend->mmvar_len =
	      sizeof(FT_MM_Var) +
	      axis_cnt*sizeof(FT_Var_Axis) +
	      instance_cnt*sizeof(FT_Var_Named_Style) +
	      instance_cnt*axis_cnt*sizeof(FT_Fixed) +
	      5 * axis_cnt;
      if ( FT_ALLOC( mmvar, face->blend->mmvar_len ))
        goto FExit;
      face->blend->mmvar = mmvar;

      mmvar->num_axis        = axis_cnt;
      mmvar->num_designs     = -1; /* Meaningless in this context */
                                   /* Each glyph may have a different number  */
			           /* of designs (or tuples, as apple calls'em*/
      mmvar->num_namedstyles = instance_cnt;
      mmvar->axis            = (FT_Var_Axis *) &(mmvar[1]);
      mmvar->namedstyle      = (FT_Var_Named_Style  *) &(mmvar->axis[axis_cnt]);
      next_coords            = (FT_Fixed *) &(mmvar->namedstyle[instance_cnt]);
      for ( i=0; i<instance_cnt; ++i )
      {
        mmvar->namedstyle[i].coords = next_coords;
	next_coords                 += axis_cnt;
      }
      next_name               = (FT_String *) next_coords;
      for ( i=0; i<axis_cnt; ++i )
      {
        mmvar->axis[i].name   = next_name;
	next_name            += 5;
      }

      FT_Stream_SeekSet( stream, fvar_start+data_off );
      a = mmvar->axis;
      for ( i=0; i<axis_cnt; ++i )
      {
	a->tag     = FT_GET_ULONG();
	a->minimum = FT_GET_ULONG();     /* A Fixed */
	a->def     = FT_GET_ULONG();     /* A Fixed */
	a->maximum = FT_GET_ULONG();     /* A Fixed */
	(void) /* flags = */ FT_GET_USHORT();
	a->strid   = FT_GET_USHORT();

	a->name[0] = a->tag>>24;
	a->name[1] = (a->tag>>16) & 0xff;
	a->name[2] = (a->tag>>8 ) & 0xff;
	a->name[3] = (a->tag    ) & 0xff;
	a->name[4] = 0;

	++a;
      }

      ns = mmvar->namedstyle;
      for ( i=0; i<instance_cnt; ++i )
      {
        ns->strid = FT_GET_USHORT();
	(void) /* flags = */ FT_GET_USHORT();
	for ( j=0; j<axis_cnt; ++j )
	  ns->coords[j] = FT_GET_ULONG();     /* A Fixed */
      }
    }

    if ( master != NULL )
    {
      if ( FT_ALLOC( mmvar, face->blend->mmvar_len ))
        goto FExit;
      FT_MEM_COPY(mmvar, face->blend->mmvar, face->blend->mmvar_len );
      mmvar->axis            = (FT_Var_Axis *) &(mmvar[1]);
      mmvar->namedstyle      = (FT_Var_Named_Style  *) &(mmvar->axis[axis_cnt]);
      next_coords            = (FT_Fixed *) &(mmvar->namedstyle[instance_cnt]);
      for ( i=0; i<instance_cnt; ++i )
      {
        mmvar->namedstyle[i].coords = next_coords;
	next_coords                 += axis_cnt;
      }
      a = mmvar->axis;
      next_name               = (FT_String *) next_coords;
      for ( i=0; i<axis_cnt; ++i )
      {
        a->name   = next_name;
	/* Standard PostScript names for some standard apple tags */
	if ( a->tag == FT_MAKE_TAG( 'w', 'g', 'h', 't' ) )
	  a->name = "Weight";
	else if ( a->tag == FT_MAKE_TAG( 'w', 'd', 't', 'h' ) )
	  a->name = "Width";
	else if ( a->tag == FT_MAKE_TAG( 'o', 'p', 's', 'z' ) )
	  a->name = "OpticalSize";
	else if ( a->tag == FT_MAKE_TAG( 's', 'l', 'n', 't' ) )
	  a->name = "Slant";
	next_name            += 5;
	++a;
      }
      *master = mmvar;
    }

  FExit:
    FT_FRAME_EXIT();
  Exit:
    return error;
  }


  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    TT_Set_MM_Blend                                                    */
  /*                                                                       */
  /* <Description>                                                         */
  /*    Sets the blend (normalized) coordinates for this instance of the   */
  /*    font. Checks that the 'gvar' table is reasonable and does some     */
  /*    initial preparation                                                */
  /*                                                                       */
  /* <InOut>                                                               */
  /*    face         :: the font.                                          */
  /*                    initialize the blend struct with gvar data         */
  /*                                                                       */
  /* <Input>                                                               */
  /*    num_coords   :: must be the axis count of the font                 */
  /*    coords       :: array of num_coords each between [-1,1]            */
  /*                                                                       */
  /* <Return>                                                              */
  /*    FreeType error code.  0 means success.                             */
  /*                                                                       */
  FT_LOCAL_DEF( FT_Error )
  TT_Set_MM_Blend( TT_Face    face,
                   FT_UInt    num_coords,
                   FT_Fixed*  coords )
  {
    FT_Error   error = TT_Err_Ok;
    GX_Blend   blend;
    FT_MM_Var *mmvar;
    FT_Int     i;
    FT_Memory  memory = face->root.memory;

    face->doblend = FALSE;

    if ( face->blend == NULL )
    {
      if ( (error = TT_Get_MM_Var( face, NULL)))
        goto Exit;
    }

    blend = face->blend;
    mmvar = blend->mmvar;

    if ( num_coords != mmvar->num_axis )
    {
      error = TT_Err_Invalid_Argument;
      goto Exit;
    }

    for ( i=0; i<num_coords; ++i )
      if ( coords[i]<-0x00010000L || coords[i]>0x00010000L )
      {
        error = TT_Err_Invalid_Argument;
        goto Exit;
      }

    if ( blend->glyphoffsets==NULL )
      if ( (error = ft_var_load_gvar(face)) )
        goto Exit;

    if ( blend->normalizedcoords == NULL )
      if ( FT_NEW_ARRAY( blend->normalizedcoords, num_coords ))
        goto Exit;

    blend->num_axis = num_coords;
    FT_MEM_COPY(blend->normalizedcoords, coords, num_coords*sizeof( FT_Fixed ));

    face->doblend = TRUE;

    if ( face->cvt != NULL )
    {
      /* The cvt table has been loaded already, every time we change the blend*/
      /*  we must reload and remodify the cvt table.                          */
      FT_FREE(face->cvt);
      face->cvt = NULL;
      tt_face_load_cvt( face, face->root.stream );
    }

  Exit:
    return error;
  }


  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    TT_Set_Var_Design                                                  */
  /*                                                                       */
  /* <Description>                                                         */
  /*    Sets the coordinates for the instance, measured in the user        */
  /*    coordinate system. Parses the 'avar' table (if present) to         */
  /*    convert from user to normalized coords                             */
  /*                                                                       */
  /* <InOut>                                                               */
  /*    face         :: the font.                                          */
  /*                    initialize the blend struct with gvar data         */
  /*                                                                       */
  /* <Input>                                                               */
  /*    num_coords   :: must be the axis count of the font                 */
  /*    coords       :: array of num_coords                                */
  /*                                                                       */
  /* <Return>                                                              */
  /*    FreeType error code.  0 means success.                             */
  /*                                                                       */
  FT_LOCAL_DEF( FT_Error )
  TT_Set_Var_Design( TT_Face   face,
                     FT_UInt   num_coords,
                     FT_Fixed* coords )
  {
    FT_Error     error = TT_Err_Ok;
    FT_Fixed    *normalized = NULL;
    GX_Blend     blend;
    FT_MM_Var   *mmvar;
    FT_Int       i,j;
    FT_Var_Axis *a;
    GX_AVar      av;
    FT_Memory    memory = face->root.memory;

    if ( face->blend == NULL )
    {
      if ( (error = TT_Get_MM_Var( face, NULL)))
        goto Exit;
    }

    blend = face->blend;
    mmvar = blend->mmvar;

    if ( num_coords != mmvar->num_axis )
    {
      error = TT_Err_Invalid_Argument;
      goto Exit;
    }

    /* Axis normalization is a two stage process. First we normalize  */
    /*  based on the [min,def,max] values for the axis to be [-1,0,1] */
    /* Then if there's an 'avar' table we renormalize this range      */
    if ( FT_NEW_ARRAY( normalized, mmvar->num_axis ))
      goto Exit;
    a = mmvar->axis;
    for ( i=0; i<mmvar->num_axis; ++i, ++a )
    {
      if ( coords[i] > a->maximum || coords[i] < a->minimum )
      {
        error = TT_Err_Invalid_Argument;
	goto Exit;
      }

      if ( coords[i] < a->def )
      {
        normalized[i] = -FT_MulDiv(
		    coords[i]-a->def,
		    0x10000L,
		    a->minimum-a->def);
      }
      else if ( a->maximum == a->def )
        normalized[i] = 0;
      else
      {
        normalized[i] = FT_MulDiv(
		    coords[i]-a->def,
		    0x10000L,
		    a->maximum-a->def);
      }
    }

    if ( !blend->avar_checked )
      ft_var_load_avar(face);

    if ( blend->avar!=NULL )
    {
      av = blend->avar;
      for ( i=0; i<mmvar->num_axis; ++i, ++av )
      {
        for ( j=1; j<av->num_points; ++j )
	  if ( normalized[i]<av->orig[j] )
	  {
	    normalized[i] = FT_MulDiv(
		    FT_MulDiv(
			normalized[i]-av->orig[j-1],
			0x10000L,
			av->orig[j]-av->orig[j-1]),
		    av->final[j]-av->final[j-1],
		    0x10000L) +
		av->final[j-1];
	break;
	  }
      }
    }

    error = TT_Set_MM_Blend( face, num_coords, normalized );

  Exit:
    FT_FREE(normalized);
    return error;
  }


  /*************************************************************************/
  /*************************************************************************/
  /*****                                                               *****/
  /*****                     GX VAR PARSING ROUTINES                   *****/
  /*****                                                               *****/
  /*************************************************************************/
  /*************************************************************************/


  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    tt_face_vary_cvt                                                   */
  /*                                                                       */
  /* <Description>                                                         */
  /*    Modifies the loaded cvt table according to the 'cvar' table and    */
  /*    the font's blend                                                   */
  /*                                                                       */
  /* <InOut>                                                               */
  /*    face   :: A handle to the target face object.                      */
  /*                                                                       */
  /* <Input>                                                               */
  /*    stream :: A handle to the input stream.                            */
  /*                                                                       */
  /* <Return>                                                              */
  /*    FreeType error code.  0 means success.                             */
  /*    Most errors I ignore.  It is perfectly legal not to have a cvar    */
  /*    table even if there is a gvar/fvar.                                */
  /*                                                                       */
  FT_LOCAL_DEF( FT_Error )
  tt_face_vary_cvt( TT_Face    face,
                    FT_Stream  stream )
  {
    FT_Error   error;
    FT_Memory  memory = stream->memory;
    FT_ULong   table_start;
    FT_ULong   table_len;
    FT_Int     tupleCount;
    FT_ULong   data_pos;
    FT_ULong   here;
    FT_Int     i,j;
    FT_Fixed  *tuple_coords = NULL;
    FT_Fixed  *im_start_coords = NULL;
    FT_Fixed  *im_end_coords = NULL;
    GX_Blend   blend = face->blend;
    FT_Int     point_count;
    FT_UShort *localpoints;
    FT_Short  *deltas;


    FT_TRACE2(( "CVAR " ));

    if ( blend==NULL )
    {
      FT_TRACE2(( "no blend specified!\n" ));

      error          = TT_Err_Ok;

      goto Exit;
    }

    if ( face->cvt==NULL )
    {
      FT_TRACE2(( "no 'cvt ' table!\n" ));

      error          = TT_Err_Ok;

      goto Exit;
    }

    error = face->goto_table( face, TTAG_cvar, stream, &table_len );
    if ( error )
    {
      FT_TRACE2(( "is missing!\n" ));

      error          = TT_Err_Ok;

      goto Exit;
    }
    if ( FT_FRAME_ENTER( table_len ) )
    {
      error          = TT_Err_Ok;
      goto Exit;
    }

    table_start = FT_Stream_FTell( stream );
    if ( FT_GET_LONG()!=0x00010000L )
    {
      FT_TRACE2(( "bad table version!\n" ));

      error          = TT_Err_Ok;

      goto FExit;
    }

    if ( FT_NEW_ARRAY( tuple_coords, blend->num_axis ) ||
        FT_NEW_ARRAY( im_start_coords, blend->num_axis ) ||
        FT_NEW_ARRAY( im_end_coords, blend->num_axis ) )
      goto FExit;

    tupleCount = FT_GET_USHORT();
    data_pos = table_start + FT_GET_USHORT();
    /* The documentation implies there are flags packed into the tuplecount */
    /*  but John Jenkins tells me that shared points don't apply to cvar */
    /*  and no other flags are defined */

    for ( i=0; i<(tupleCount&0xfff); ++i ) {
      FT_Int   tupleDataSize;
      FT_Int   tupleIndex;
      FT_Fixed apply;
      tupleDataSize = FT_GET_USHORT();
      tupleIndex = FT_GET_USHORT();
      /* there is no provision here for a global tuple coordinate section */
      /*  so John says there are no tuple indeces. Just embedded tuples */
      if ( tupleIndex & 0x8000 )
      {
        for ( j=0; j<blend->num_axis; ++j )
          tuple_coords[j] = FT_GET_SHORT()<<2;	/* Convert from short frac to fixed */
      }
      else
      {
        /* Skip this tuple, it makes no sense */
        if ( tupleIndex&0x4000 )
          for ( j=0; j<2*blend->num_axis; ++j )
	    (void) FT_GET_SHORT();
        data_pos += tupleDataSize;
    continue;
      }
      if ( tupleIndex & 0x4000 )
      {
        for ( j=0; j<blend->num_axis; ++j )
          im_start_coords[j] = FT_GET_SHORT()<<2;
        for ( j=0; j<blend->num_axis; ++j )
          im_end_coords[j] = FT_GET_SHORT()<<2;
      }
      apply = ft_var_apply_tuple(blend,
		    tupleIndex,
		    tuple_coords,
		    im_start_coords,
		    im_end_coords);
      if ( apply==0 ||             /* tuple isn't active for our blend */
          !(tupleIndex&0x2000) )   /* global points not allowed, if they aren't local, makes no sense */
      {
        data_pos += tupleDataSize;
    continue;
      }

      here = FT_Stream_FTell(stream);

      FT_Stream_SeekSet( stream, data_pos );
      localpoints = ft_var_readpackedpoints( stream, &point_count );
      deltas = ft_var_readpackeddeltas( stream,
	        point_count==0 ? face->cvt_size : point_count );
      if ( localpoints==NULL || deltas == NULL )
        /* Failure, ignore it */;
      else if ( localpoints == ALL_POINTS )
      {      /* Means that there are deltas for every entry in cvt */
        for ( j=0; j<face->cvt_size; ++j )
          face->cvt[j] += FT_MulFix(deltas[j], apply);
      }
      else
      {
        for ( j=0; j<point_count; ++j )
          face->cvt[ localpoints[j] ] += FT_MulFix(deltas[j], apply);
      }
      if ( localpoints!=ALL_POINTS )
        FT_FREE( localpoints );
      FT_FREE( deltas );

      data_pos += tupleDataSize;
      FT_Stream_SeekSet( stream, here );
    }

  FExit:
    FT_FRAME_EXIT();
  Exit:
    FT_FREE(tuple_coords);
    FT_FREE(im_start_coords);
    FT_FREE(im_end_coords);

    return error;
  }



  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    TT_Vary_Get_Glyph_Deltas                                           */
  /*                                                                       */
  /* <Description>                                                         */
  /*    Loads the appropriate deltas for the current glyph                 */
  /*                                                                       */
  /* <Input>                                                               */
  /*    face        :: A handle to the target face object.                 */
  /*                                                                       */
  /*    glyph_index :: The number of the glyph being modified.             */
  /*                                                                       */
  /*    n_points    :: The number of the points in the glyph.              */
  /*                   includes phantom points                             */
  /*                                                                       */
  /* <Output>                                                              */
  /*    deltas      :: The array of points to change.                      */
  /*                                                                       */
  /* <Return>                                                              */
  /*    FreeType error code.  0 means success.                             */
  /*                                                                       */
  FT_LOCAL_DEF( FT_Error )
  TT_Vary_Get_Glyph_Deltas( TT_Face    face,
			    FT_UInt    glyph_index,
		       	    FT_Vector  **_delta_xy,
			    FT_UInt    n_points)
  {
    FT_Stream  stream = face->root.stream;
    FT_Memory  memory = stream->memory;
    GX_Blend   blend = face->blend;
    FT_Vector *delta_xy;

    FT_Error   error;
    FT_ULong   glyph_start;
    FT_Int     tupleCount;
    FT_ULong   data_pos;
    FT_ULong   here;
    FT_Int     i,j;
    FT_Fixed  *tuple_coords = NULL;
    FT_Fixed  *im_start_coords = NULL;
    FT_Fixed  *im_end_coords = NULL;
    FT_Int     point_count, spoint_count = 0;
    FT_UShort *sharedpoints = NULL;
    FT_UShort *localpoints;
    FT_UShort *points;
    FT_Short  *deltas_x, *deltas_y;

    if ( !face->doblend || blend==NULL )
      return TT_Err_Invalid_Argument;

    if ( (error = FT_NEW_ARRAY(delta_xy, n_points)) )
      goto Fail;
    FT_MEM_ZERO( delta_xy, n_points*sizeof(FT_Vector));
    *_delta_xy = delta_xy;

    if ( glyph_index >= blend->gv_glyphcnt ||
	blend->glyphoffsets[glyph_index] == blend->glyphoffsets[glyph_index+1] )
      return TT_Err_Ok;               /* No variation data for this glyph */

    if ( FT_STREAM_SEEK( blend->glyphoffsets[glyph_index] ) ||
         FT_FRAME_ENTER(blend->glyphoffsets[glyph_index+1] - blend->glyphoffsets[glyph_index]))
      goto Fail;

    glyph_start = FT_Stream_FTell( stream );


    /* The each set of glyph variation data is formatted similarly to cvar */
    /* (except we get shared points and global tuples)                     */

    if ( FT_NEW_ARRAY( tuple_coords, blend->num_axis ) ||
        FT_NEW_ARRAY( im_start_coords, blend->num_axis ) ||
        FT_NEW_ARRAY( im_end_coords, blend->num_axis ) )
      goto Exit;

    tupleCount = FT_GET_USHORT();
    data_pos   = glyph_start + FT_GET_USHORT();
    if ( tupleCount & 0x8000 )
    {
      FT_ULong here = FT_Stream_FTell( stream );
      FT_Stream_SeekSet( stream, data_pos );
      sharedpoints = ft_var_readpackedpoints( stream, &spoint_count );
      data_pos = FT_Stream_FTell( stream );
      FT_Stream_SeekSet( stream, here );
    }

    for ( i=0; i<(tupleCount&0xfff); ++i ) {
      FT_Int   tupleDataSize;
      FT_Int   tupleIndex;
      FT_Fixed apply;
      tupleDataSize = FT_GET_USHORT();
      tupleIndex = FT_GET_USHORT();

      if ( tupleIndex & 0x8000 )
      {
        for ( j=0; j<blend->num_axis; ++j )
          tuple_coords[j] = FT_GET_SHORT()<<2;	/* Convert from short frac to fixed */
      }
      else if ( (tupleIndex&0xfff) >= blend->tuplecount )
      {
	error = TT_Err_Invalid_Table;
	goto Fail;
      }
      else
      {
        FT_MEM_COPY( tuple_coords, &blend->tuplecoords[(tupleIndex&0xfff)*blend->num_axis],
	    blend->num_axis*sizeof(FT_Fixed));
      }
      if ( tupleIndex & 0x4000 )
      {
        for ( j=0; j<blend->num_axis; ++j )
          im_start_coords[j] = FT_GET_SHORT()<<2;
        for ( j=0; j<blend->num_axis; ++j )
          im_end_coords[j] = FT_GET_SHORT()<<2;
      }
      apply = ft_var_apply_tuple(blend,
		    tupleIndex,
		    tuple_coords,
		    im_start_coords,
		    im_end_coords);
      if ( apply==0 )              /* tuple isn't active for our blend */
      {
        data_pos += tupleDataSize;
    continue;
      }

      here = FT_Stream_FTell(stream);
      if ( tupleIndex&0x2000 )
      {
        FT_Stream_SeekSet( stream, data_pos );
        localpoints = ft_var_readpackedpoints( stream, &point_count );
	points = localpoints;
      }
      else
      {
        points = sharedpoints;
	point_count = spoint_count;
      }
      deltas_x = ft_var_readpackeddeltas( stream,
	        point_count==0 ? n_points : point_count );
      deltas_y = ft_var_readpackeddeltas( stream,
	        point_count==0 ? n_points : point_count );
      if ( points==NULL || deltas_y == NULL || deltas_x == NULL )
        /* Failure, ignore it */;
      else if ( points == ALL_POINTS )
      {      /* Means that there are deltas for every point in the glyph */
        for ( j=0; j<n_points; ++j )
	{
          delta_xy[j].x += FT_MulFix(deltas_x[j], apply);
          delta_xy[j].y += FT_MulFix(deltas_y[j], apply);
        }
      }
      else
      {
        for ( j=0; j<point_count; ++j )
	{
          delta_xy[ localpoints[j] ].x += FT_MulFix(deltas_x[j], apply);
          delta_xy[ localpoints[j] ].y += FT_MulFix(deltas_y[j], apply);
        }
      }
      if ( localpoints!=ALL_POINTS )
        FT_FREE( localpoints );
      FT_FREE( deltas_x );
      FT_FREE( deltas_y );

      data_pos += tupleDataSize;
      FT_Stream_SeekSet( stream, here );
    }

    goto Exit;

  Fail:
    FT_FREE(delta_xy);
    *_delta_xy = NULL;

  Exit:
    FT_FREE(tuple_coords);
    FT_FREE(im_start_coords);
    FT_FREE(im_end_coords);

    return error;
  }


  /*************************************************************************/
  /*                                                                       */
  /* <Function>                                                            */
  /*    tt_done_blend                                                      */
  /*                                                                       */
  /* <Description>                                                         */
  /*    frees the blend internal data structure                            */
  /*                                                                       */
  FT_LOCAL_DEF( void )
  tt_done_blend( FT_Memory memory,
      GX_Blend    blend )
  {
    if ( blend!=NULL )
    {
      FT_Int i;

      FT_FREE(blend->normalizedcoords);
      FT_FREE(blend->mmvar);
      for ( i=0; i<blend->num_axis; ++i )
      {
        FT_FREE(blend->avar[i].orig);
        FT_FREE(blend->avar[i].final);
      }
      FT_FREE(blend->avar);
      FT_FREE(blend->tuplecoords);
      FT_FREE(blend->glyphoffsets);
      FT_FREE(blend);
    }
  }
#endif
